# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Integration tests for various transaction settlement skill's failure modes."""


import binascii
import os
import tempfile
from abc import ABC
from math import ceil
from typing import Any, Dict, Union, cast

from aea.crypto.base import Crypto
from aea.crypto.registries import make_crypto, make_ledger_api
from aea_ledger_ethereum import EthereumApi
from aea_test_autonomy.helpers.contracts import get_register_contract
from web3.types import Nonce, Wei

from packages.open_aea.protocols.signing import SigningMessage
from packages.valory.contracts.gnosis_safe.tests.test_contract import (
    PACKAGE_DIR as GNOSIS_SAFE_PACKAGE,
)
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.contract_api.custom_types import RawTransaction, State
from packages.valory.protocols.ledger_api import LedgerApiMessage
from packages.valory.protocols.ledger_api.custom_types import (
    SignedTransaction,
    TransactionDigest,
    TransactionReceipt,
)
from packages.valory.skills.abstract_round_abci.test_tools.integration import (
    ExpectedContentType,
    ExpectedTypesType,
    HandlersType,
    IntegrationBaseCase,
)
from packages.valory.skills.transaction_settlement_abci.behaviours import (
    FinalizeBehaviour,
    ValidateTransactionBehaviour,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    VerificationStatus,
    skill_input_hex_to_payload,
)
from packages.valory.skills.transaction_settlement_abci.payloads import SignaturePayload
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TxSettlementSynchronizedSata,
)


# pylint: disable=protected-access,too-many-ancestors,unbalanced-tuple-unpacking,too-many-locals,consider-using-with,unspecified-encoding,too-many-arguments,unidiomatic-typecheck


DUMMY_MAX_FEE_PER_GAS = 4000000000
DUMMY_MAX_PRIORITY_FEE_PER_GAS = 3000000000
DUMMY_REPRICING_MULTIPLIER = 1.1


class _SafeConfiguredHelperIntegration(IntegrationBaseCase, ABC):  # pragma: no cover
    """Base test class for integration tests with Gnosis, but no contract, deployed."""

    safe_owners: Dict[str, Crypto]
    keeper_address: str

    @classmethod
    def setup_class(cls, **kwargs: Any) -> None:
        """Setup."""
        super().setup_class()

        # safe configuration
        cls.safe_owners = {}
        for address, p_key in cls.agents.items():
            with tempfile.TemporaryDirectory() as temp_dir:
                fp = os.path.join(temp_dir, "key.txt")
                f = open(fp, "w")
                f.write(p_key)
                f.close()
                crypto = make_crypto("ethereum", private_key_path=str(fp))
            cls.safe_owners[address] = crypto
        cls.keeper_address = cls.current_agent
        assert cls.keeper_address in cls.safe_owners  # nosec


class _GnosisHelperIntegration(
    _SafeConfiguredHelperIntegration, ABC
):  # pragma: no cover
    """Class that assists Gnosis instantiation."""

    safe_contract_address: str = "0x68FCdF52066CcE5612827E872c45767E5a1f6551"
    ethereum_api: EthereumApi
    gnosis_instance: Any

    @classmethod
    def setup_class(cls, **kwargs: Any) -> None:
        """Setup."""
        super().setup_class()

        # register gnosis contract
        gnosis = get_register_contract(GNOSIS_SAFE_PACKAGE)

        cls.ethereum_api = make_ledger_api("ethereum")
        cls.gnosis_instance = gnosis.get_instance(
            cls.ethereum_api, cls.safe_contract_address
        )


class _TxHelperIntegration(_GnosisHelperIntegration, ABC):  # pragma: no cover
    """Class that assists tx settlement related operations."""

    tx_settlement_synchronized_data: TxSettlementSynchronizedSata

    def sign_tx(self) -> None:
        """Sign a transaction"""
        tx_params = skill_input_hex_to_payload(
            self.tx_settlement_synchronized_data.most_voted_tx_hash
        )
        safe_tx_hash_bytes = binascii.unhexlify(tx_params["safe_tx_hash"])
        participant_to_signature = {}
        for address, crypto in self.safe_owners.items():
            signature_hex = crypto.sign_message(
                safe_tx_hash_bytes,
                is_deprecated_mode=True,
            )
            signature_hex = signature_hex[2:]
            participant_to_signature[address] = SignaturePayload(
                sender=address,
                signature=signature_hex,
            ).json

        # FIXME: The following loop is a patch. The
        #  [_get_python_modules](https://github.com/valory-xyz/open-aea/blob/d0e60881b1371442c3572df86c53fc92dc9228fa/aea/skills/base.py#L907-L925)
        #  is getting the python modules from the skill directory.
        #  As we can see from the code, the path will end up being relative, which means that the
        #  [_metaclass_registry_key](https://github.com/valory-xyz/open-autonomy/blob/5d151f1fff4934f70be8c5f6be77705cc2e6ef4c/packages/valory/skills/abstract_round_abci/base.py#L167)
        #  inserted in the `_MetaPayload`'s registry will also be relative. However, this is causing issues when calling
        #  `BaseTxPayload.from_json(payload_json)` later from the property below
        #  (`self.tx_settlement_synchronized_data.participant_to_signature`) because the `payload_json` will have been
        #  serialized using an imported payload (the `SignaturePayload` above), and therefore a key error will be
        #  raised since the imported payload's path is not relative and the registry has a relative path as a key.
        for payload in participant_to_signature.values():
            registry_key = "_metaclass_registry_key"
            payload_value = payload[registry_key]
            payload_cls_name = payload_value.split(".")[-1]
            patched_registry_key = f"payloads.{payload_cls_name}"
            payload[registry_key] = patched_registry_key

        self.tx_settlement_synchronized_data.update(
            participant_to_signature=participant_to_signature,
        )

        actual_safe_owners = self.gnosis_instance.functions.getOwners().call()
        expected_safe_owners = (
            self.tx_settlement_synchronized_data.participant_to_signature.keys()
        )
        assert len(actual_safe_owners) == len(expected_safe_owners)  # nosec
        assert all(  # nosec
            owner == signer
            for owner, signer in zip(actual_safe_owners, expected_safe_owners)
        )

    def send_tx(self, simulate_timeout: bool = False) -> None:
        """Send a transaction"""

        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=FinalizeBehaviour.auto_behaviour_id(),
            synchronized_data=self.tx_settlement_synchronized_data,
        )
        behaviour = cast(FinalizeBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == FinalizeBehaviour.auto_behaviour_id()
        stored_nonce = behaviour.params.mutable_params.nonce
        stored_gas_price = behaviour.params.mutable_params.gas_price

        handlers: HandlersType = [
            self.contract_handler,
            self.signing_handler,
            self.ledger_handler,
        ]
        expected_content: ExpectedContentType = [
            {"performative": ContractApiMessage.Performative.RAW_TRANSACTION},
            {"performative": SigningMessage.Performative.SIGNED_TRANSACTION},
            {"performative": LedgerApiMessage.Performative.TRANSACTION_DIGEST},
        ]
        expected_types: ExpectedTypesType = [
            {
                "raw_transaction": RawTransaction,
            },
            {
                "signed_transaction": SignedTransaction,
            },
            {
                "transaction_digest": TransactionDigest,
            },
        ]
        msg1, _, msg3 = self.process_n_messages(
            3,
            self.tx_settlement_synchronized_data,
            None,
            handlers,
            expected_content,
            expected_types,
            fail_send_a2a=simulate_timeout,
        )
        assert msg1 is not None and isinstance(msg1, ContractApiMessage)  # nosec
        assert msg3 is not None and isinstance(msg3, LedgerApiMessage)  # nosec
        nonce_used = Nonce(int(cast(str, msg1.raw_transaction.body["nonce"])))
        gas_price_used = {
            gas_price_param: Wei(
                int(
                    cast(
                        str,
                        msg1.raw_transaction.body[gas_price_param],
                    )
                )
            )
            for gas_price_param in ("maxPriorityFeePerGas", "maxFeePerGas")
        }
        tx_digest = msg3.transaction_digest.body
        tx_data = {
            "status": VerificationStatus.PENDING,
            "tx_digest": cast(str, tx_digest),
        }

        behaviour = cast(FinalizeBehaviour, self.behaviour.current_behaviour)
        assert behaviour.params.mutable_params.gas_price == gas_price_used  # nosec
        assert behaviour.params.mutable_params.nonce == nonce_used  # nosec
        if simulate_timeout:
            assert behaviour.params.mutable_params.tx_hash == tx_digest  # nosec
        else:
            assert behaviour.params.mutable_params.tx_hash == ""  # nosec

        # if we are repricing
        if nonce_used == stored_nonce:
            assert stored_nonce is not None  # nosec
            assert stored_gas_price is not None  # nosec
            assert gas_price_used == {  # nosec
                gas_price_param: ceil(
                    stored_gas_price[gas_price_param] * DUMMY_REPRICING_MULTIPLIER
                )
                for gas_price_param in ("maxPriorityFeePerGas", "maxFeePerGas")
            }, "The repriced parameters do not match the ones returned from the gas pricing method!"
        # if we are not repricing
        else:
            assert gas_price_used == {  # nosec
                "maxPriorityFeePerGas": DUMMY_MAX_PRIORITY_FEE_PER_GAS,
                "maxFeePerGas": DUMMY_MAX_FEE_PER_GAS,
            }, "The used parameters do not match the ones returned from the gas pricing method!"

        update_params: Dict[str, Union[int, str, Dict[str, int]]]
        if not simulate_timeout:
            hashes = self.tx_settlement_synchronized_data.tx_hashes_history
            hashes.append(tx_digest)
            update_params = dict(
                tx_hashes_history="".join(hashes),
                final_verification_status=VerificationStatus(tx_data["status"]).value,
            )
        else:
            # store the tx hash that we have missed and update missed messages.
            assert isinstance(  # nosec
                self.behaviour.current_behaviour, FinalizeBehaviour
            )
            self.mock_a2a_transaction()
            self.behaviour.current_behaviour.params.mutable_params.tx_hash = tx_digest
            missed_messages = self.tx_settlement_synchronized_data.missed_messages
            missed_messages[
                self.tx_settlement_synchronized_data.most_voted_keeper_address
            ] += 1
            update_params = dict(missed_messages=missed_messages)

        self.tx_settlement_synchronized_data.update(
            synchronized_data_class=None, **update_params
        )

    def validate_tx(
        self, simulate_timeout: bool = False, mining_interval_secs: float = 0
    ) -> None:
        """Validate the sent transaction."""

        if simulate_timeout:
            missed_messages = self.tx_settlement_synchronized_data.missed_messages
            missed_messages[
                tuple(self.tx_settlement_synchronized_data.all_participants)[0]
            ] += 1
            self.tx_settlement_synchronized_data.update(missed_messages=missed_messages)
        else:
            handlers: HandlersType = [
                self.ledger_handler,
                self.contract_handler,
            ]
            expected_content: ExpectedContentType = [
                {"performative": LedgerApiMessage.Performative.TRANSACTION_RECEIPT},
                {"performative": ContractApiMessage.Performative.STATE},
            ]
            expected_types: ExpectedTypesType = [
                {
                    "transaction_receipt": TransactionReceipt,
                },
                {
                    "state": State,
                },
            ]
            _, verif_msg = self.process_n_messages(
                2,
                self.tx_settlement_synchronized_data,
                ValidateTransactionBehaviour.auto_behaviour_id(),
                handlers,
                expected_content,
                expected_types,
                mining_interval_secs=mining_interval_secs,
            )
            assert verif_msg is not None and isinstance(  # nosec
                verif_msg, ContractApiMessage
            )
            assert verif_msg.state.body[  # nosec
                "verified"
            ], f"Message not verified: {verif_msg.state.body}"

            self.tx_settlement_synchronized_data.update(
                final_verification_status=VerificationStatus.VERIFIED.value,
                final_tx_hash=self.tx_settlement_synchronized_data.to_be_validated_tx_hash,
            )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ipfs.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nipfs.proto\x12\x16\x61\x65\x61.valory.ipfs.v0_1_0"\xb2\x07\n\x0bIpfsMessage\x12G\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x36.aea.valory.ipfs.v0_1_0.IpfsMessage.Error_PerformativeH\x00\x12G\n\x05\x66iles\x18\x06 \x01(\x0b\x32\x36.aea.valory.ipfs.v0_1_0.IpfsMessage.Files_PerformativeH\x00\x12O\n\tget_files\x18\x07 \x01(\x0b\x32:.aea.valory.ipfs.v0_1_0.IpfsMessage.Get_Files_PerformativeH\x00\x12O\n\tipfs_hash\x18\x08 \x01(\x0b\x32:.aea.valory.ipfs.v0_1_0.IpfsMessage.Ipfs_Hash_PerformativeH\x00\x12S\n\x0bstore_files\x18\t \x01(\x0b\x32<.aea.valory.ipfs.v0_1_0.IpfsMessage.Store_Files_PerformativeH\x00\x1a\xc9\x01\n\x18Store_Files_Performative\x12V\n\x05\x66iles\x18\x01 \x03(\x0b\x32G.aea.valory.ipfs.v0_1_0.IpfsMessage.Store_Files_Performative.FilesEntry\x12\x0f\n\x07timeout\x18\x02 \x01(\x01\x12\x16\n\x0etimeout_is_set\x18\x03 \x01(\x08\x1a,\n\nFilesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a+\n\x16Ipfs_Hash_Performative\x12\x11\n\tipfs_hash\x18\x01 \x01(\t\x1aT\n\x16Get_Files_Performative\x12\x11\n\tipfs_hash\x18\x01 \x01(\t\x12\x0f\n\x07timeout\x18\x02 \x01(\x01\x12\x16\n\x0etimeout_is_set\x18\x03 \x01(\x08\x1a\x94\x01\n\x12\x46iles_Performative\x12P\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x41.aea.valory.ipfs.v0_1_0.IpfsMessage.Files_Performative.FilesEntry\x1a,\n\nFilesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a$\n\x12\x45rror_Performative\x12\x0e\n\x06reason\x18\x01 \x01(\tB\x0e\n\x0cperformativeb\x06proto3'
)


_IPFSMESSAGE = DESCRIPTOR.message_types_by_name["IpfsMessage"]
_IPFSMESSAGE_STORE_FILES_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Store_Files_Performative"
]
_IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY = (
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE.nested_types_by_name["FilesEntry"]
)
_IPFSMESSAGE_IPFS_HASH_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Ipfs_Hash_Performative"
]
_IPFSMESSAGE_GET_FILES_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Get_Files_Performative"
]
_IPFSMESSAGE_FILES_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Files_Performative"
]
_IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY = (
    _IPFSMESSAGE_FILES_PERFORMATIVE.nested_types_by_name["FilesEntry"]
)
_IPFSMESSAGE_ERROR_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Error_Performative"
]
IpfsMessage = _reflection.GeneratedProtocolMessageType(
    "IpfsMessage",
    (_message.Message,),
    {
        "Store_Files_Performative": _reflection.GeneratedProtocolMessageType(
            "Store_Files_Performative",
            (_message.Message,),
            {
                "FilesEntry": _reflection.GeneratedProtocolMessageType(
                    "FilesEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY,
                        "__module__": "ipfs_pb2"
                        # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Store_Files_Performative.FilesEntry)
                    },
                ),
                "DESCRIPTOR": _IPFSMESSAGE_STORE_FILES_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Store_Files_Performative)
            },
        ),
        "Ipfs_Hash_Performative": _reflection.GeneratedProtocolMessageType(
            "Ipfs_Hash_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _IPFSMESSAGE_IPFS_HASH_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Ipfs_Hash_Performative)
            },
        ),
        "Get_Files_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Files_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _IPFSMESSAGE_GET_FILES_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Get_Files_Performative)
            },
        ),
        "Files_Performative": _reflection.GeneratedProtocolMessageType(
            "Files_Performative",
            (_message.Message,),
            {
                "FilesEntry": _reflection.GeneratedProtocolMessageType(
                    "FilesEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY,
                        "__module__": "ipfs_pb2"
                        # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Files_Performative.FilesEntry)
                    },
                ),
                "DESCRIPTOR": _IPFSMESSAGE_FILES_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Files_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _IPFSMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _IPFSMESSAGE,
        "__module__": "ipfs_pb2"
        # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage)
    },
)
_sym_db.RegisterMessage(IpfsMessage)
_sym_db.RegisterMessage(IpfsMessage.Store_Files_Performative)
_sym_db.RegisterMessage(IpfsMessage.Store_Files_Performative.FilesEntry)
_sym_db.RegisterMessage(IpfsMessage.Ipfs_Hash_Performative)
_sym_db.RegisterMessage(IpfsMessage.Get_Files_Performative)
_sym_db.RegisterMessage(IpfsMessage.Files_Performative)
_sym_db.RegisterMessage(IpfsMessage.Files_Performative.FilesEntry)
_sym_db.RegisterMessage(IpfsMessage.Error_Performative)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY._options = None
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY._serialized_options = b"8\001"
    _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY._options = None
    _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY._serialized_options = b"8\001"
    _IPFSMESSAGE._serialized_start = 39
    _IPFSMESSAGE._serialized_end = 985
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE._serialized_start = 448
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE._serialized_end = 649
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY._serialized_start = 605
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY._serialized_end = 649
    _IPFSMESSAGE_IPFS_HASH_PERFORMATIVE._serialized_start = 651
    _IPFSMESSAGE_IPFS_HASH_PERFORMATIVE._serialized_end = 694
    _IPFSMESSAGE_GET_FILES_PERFORMATIVE._serialized_start = 696
    _IPFSMESSAGE_GET_FILES_PERFORMATIVE._serialized_end = 780
    _IPFSMESSAGE_FILES_PERFORMATIVE._serialized_start = 783
    _IPFSMESSAGE_FILES_PERFORMATIVE._serialized_end = 931
    _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY._serialized_start = 605
    _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY._serialized_end = 649
    _IPFSMESSAGE_ERROR_PERFORMATIVE._serialized_start = 933
    _IPFSMESSAGE_ERROR_PERFORMATIVE._serialized_end = 969
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Mint command group definitions."""

from pathlib import Path
from typing import Optional, Tuple, Union, cast

import click
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import pass_ctx
from aea.configurations.data_types import PackageType
from aea.helpers.base import IPFSHash

from autonomy.chain.config import ChainType
from autonomy.cli.helpers.chain import mint_component, mint_service
from autonomy.cli.utils.click_utils import (
    NFTArgument,
    PathArgument,
    chain_selection_flag,
)


package_path_decorator = click.argument(
    "package_path",
    type=PathArgument(exists=True, file_okay=False, dir_okay=True),
)
password_decorator = click.option(
    "--password",
    type=str,
    help="Password for key pair",
)
dependencies_decorator = click.option(
    "-d",
    "--dependencies",
    type=str,
    multiple=True,
    help="Dependencies for the package",
)
nft_decorator = click.option(
    "--nft",
    type=NFTArgument(),
    help="IPFS hash or path for the NFT image",
)
timeout_flag = click.option(
    "-t", "--timeout", type=float, help="Timeout for verifying emnitted events"
)
owner_flag = click.option(
    "--owner",
    type=str,
    help="Owner address for the component",
)

key_path_decorator = click.option(
    "--key",
    type=PathArgument(exists=True, file_okay=True, dir_okay=False),
    help="Use private key from a file for signing the transactions",
)
hwi_flag = click.option(
    "--hwi",
    is_flag=True,
    help="Use hardware wallet for signing the transactions.",
)


@click.group("mint")
@pass_ctx
@chain_selection_flag()
@timeout_flag
@click.option(
    "--skip-hash-check",
    is_flag=True,
    help="Skip hash check when verifying dependencies on chain",
)
def mint(  # pylint: disable=too-many-arguments
    ctx: Context,
    chain_type: str,
    skip_hash_check: bool,
    timeout: float,
) -> None:
    """Mint component on-chain."""

    ctx.config["chain_type"] = ChainType(chain_type)
    ctx.config["skip_hash_check"] = skip_hash_check
    ctx.config["timeout"] = timeout


@mint.command()
@package_path_decorator
@key_path_decorator
@hwi_flag
@password_decorator
@dependencies_decorator
@nft_decorator
@owner_flag
@pass_ctx
def protocol(  # pylint: disable=too-many-arguments
    ctx: Context,
    package_path: Path,
    key: Path,
    password: Optional[str],
    dependencies: Tuple[str],
    nft: Optional[Union[Path, IPFSHash]],
    owner: Optional[str],
    hwi: bool = False,
) -> None:
    """Mint a protocol component."""

    mint_component(
        package_path=package_path,
        package_type=PackageType.PROTOCOL,
        key=key,
        chain_type=cast(ChainType, ctx.config.get("chain_type")),
        dependencies=list(map(int, dependencies)),
        password=password,
        nft=nft,
        owner=owner,
        skip_hash_check=ctx.config.get("skip_hash_check", False),
        timeout=ctx.config["timeout"],
        hwi=hwi,
    )


@mint.command()
@package_path_decorator
@key_path_decorator
@hwi_flag
@password_decorator
@dependencies_decorator
@nft_decorator
@owner_flag
@pass_ctx
def contract(  # pylint: disable=too-many-arguments
    ctx: Context,
    package_path: Path,
    key: Path,
    password: Optional[str],
    dependencies: Tuple[str],
    nft: Optional[Union[Path, IPFSHash]],
    owner: Optional[str],
    hwi: bool = False,
) -> None:
    """Mint a contract component."""

    mint_component(
        package_path=package_path,
        package_type=PackageType.CONTRACT,
        key=key,
        chain_type=cast(ChainType, ctx.config.get("chain_type")),
        dependencies=list(map(int, dependencies)),
        password=password,
        nft=nft,
        owner=owner,
        skip_hash_check=ctx.config.get("skip_hash_check", False),
        timeout=ctx.config["timeout"],
        hwi=hwi,
    )


@mint.command()
@package_path_decorator
@key_path_decorator
@hwi_flag
@password_decorator
@dependencies_decorator
@nft_decorator
@owner_flag
@pass_ctx
def connection(  # pylint: disable=too-many-arguments
    ctx: Context,
    package_path: Path,
    key: Path,
    password: Optional[str],
    dependencies: Tuple[str],
    nft: Optional[Union[Path, IPFSHash]],
    owner: Optional[str],
    hwi: bool = False,
) -> None:
    """Mint a connection component."""

    mint_component(
        package_path=package_path,
        package_type=PackageType.CONNECTION,
        key=key,
        chain_type=cast(ChainType, ctx.config.get("chain_type")),
        dependencies=list(map(int, dependencies)),
        password=password,
        nft=nft,
        owner=owner,
        skip_hash_check=ctx.config.get("skip_hash_check", False),
        timeout=ctx.config["timeout"],
        hwi=hwi,
    )


@mint.command()
@package_path_decorator
@key_path_decorator
@hwi_flag
@password_decorator
@dependencies_decorator
@nft_decorator
@owner_flag
@pass_ctx
def skill(  # pylint: disable=too-many-arguments
    ctx: Context,
    package_path: Path,
    key: Path,
    password: Optional[str],
    dependencies: Tuple[str],
    nft: Optional[Union[Path, IPFSHash]],
    owner: Optional[str],
    hwi: bool = False,
) -> None:
    """Mint a skill component."""

    mint_component(
        package_path=package_path,
        package_type=PackageType.SKILL,
        key=key,
        chain_type=cast(ChainType, ctx.config.get("chain_type")),
        dependencies=list(map(int, dependencies)),
        password=password,
        nft=nft,
        owner=owner,
        skip_hash_check=ctx.config.get("skip_hash_check", False),
        timeout=ctx.config["timeout"],
        hwi=hwi,
    )


@mint.command()
@package_path_decorator
@key_path_decorator
@hwi_flag
@password_decorator
@dependencies_decorator
@nft_decorator
@owner_flag
@pass_ctx
def agent(  # pylint: disable=too-many-arguments
    ctx: Context,
    package_path: Path,
    key: Path,
    password: Optional[str],
    dependencies: Tuple[str],
    nft: Optional[Union[Path, IPFSHash]],
    owner: Optional[str],
    hwi: bool = False,
) -> None:
    """Mint an agent."""

    mint_component(
        package_path=package_path,
        package_type=PackageType.AGENT,
        key=key,
        chain_type=cast(ChainType, ctx.config.get("chain_type")),
        dependencies=list(map(int, dependencies)),
        password=password,
        nft=nft,
        owner=owner,
        skip_hash_check=ctx.config.get("skip_hash_check", False),
        timeout=ctx.config["timeout"],
        hwi=hwi,
    )


@mint.command()
@package_path_decorator
@key_path_decorator
@hwi_flag
@password_decorator
@nft_decorator
@owner_flag
@pass_ctx
@click.option(
    "-a",
    "--agent-id",
    type=int,
    help="Canonical agent ID",
    required=True,
)
@click.option(
    "-n",
    "--number-of-slots",
    type=int,
    help="Number of agent instances for the agent",
    required=True,
)
@click.option(
    "-c",
    "--cost-of-bond",
    type=int,
    help="Cost of bond for the agent (Wei)",
    required=True,
)
@click.option(
    "--threshold",
    type=int,
    required=True,
    help="Threshold for the minimum numbers required to run the service",
)
def service(  # pylint: disable=too-many-arguments  # pylint: disable=too-many-arguments
    ctx: Context,
    package_path: Path,
    key: Path,
    agent_id: int,
    number_of_slots: int,
    cost_of_bond: int,
    threshold: int,
    password: Optional[str],
    nft: Optional[Union[Path, IPFSHash]],
    owner: Optional[str],
    hwi: bool = False,
) -> None:
    """Mint a service"""

    mint_service(
        package_path=package_path,
        key=key,
        agent_id=agent_id,
        number_of_slots=number_of_slots,
        cost_of_bond=cost_of_bond,
        threshold=threshold,
        chain_type=cast(ChainType, ctx.config.get("chain_type")),
        password=password,
        nft=nft,
        owner=owner,
        skip_hash_check=ctx.config.get("skip_hash_check", False),
        timeout=ctx.config["timeout"],
        hwi=hwi,
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test chain helpers."""

from unittest import mock

import click
import pytest
from aea.configurations.data_types import PackageType
from aea_ledger_ethereum_hwi.hwi import EthereumHWIApi, EthereumHWICrypto
from aea_test_autonomy.configurations import ETHEREUM_KEY_DEPLOYER

from autonomy.chain.base import ServiceState
from autonomy.chain.config import ChainConfigs, ChainType
from autonomy.chain.mint import registry_contracts
from autonomy.cli.helpers.chain import (
    activate_service,
    deploy_service,
    get_ledger_and_crypto_objects,
    mint_component,
    mint_service,
    register_instance,
    terminate_service,
    unbond_service,
)

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.test_mint.test_mint_components import DummyContract


PACKAGE_DIR = ROOT_DIR / "packages" / "valory" / "protocols" / "abci"
DUMMY_METADATA_HASH = (
    "0xd913b5bf68193dfacb941538d5900466c449c9ec8121153f152de2e026fa7f3a"
)


publish_metadata_patch = mock.patch(
    "autonomy.cli.helpers.chain.publish_metadata",
    return_value=(DUMMY_METADATA_HASH, ""),
)

_ = registry_contracts.service_registry
_ = registry_contracts.service_manager
_ = registry_contracts.registries_manager
_ = registry_contracts.component_registry
_ = registry_contracts.service_manager


class TestMintComponentMethod:
    """Test `mint_component` method."""

    def test_mint_component_rpc_connect_fail(
        self,
    ) -> None:
        """Test RPC connection error."""

        with pytest.raises(
            click.ClickException,
            match="Component mint failed with following error; Cannot connect to the given RPC",
        ):
            with publish_metadata_patch, mock.patch(
                "autonomy.chain.mint.Crypto.sign_transaction"
            ):
                mint_component(
                    package_path=PACKAGE_DIR,
                    package_type=PackageType.PROTOCOL,
                    key=ETHEREUM_KEY_DEPLOYER,
                    chain_type=ChainType.LOCAL,
                    dependencies=[],
                )

    def test_mint_component_token_id_retrieve_fail(
        self,
    ) -> None:
        """Test token ID retrieval failure method."""
        with pytest.raises(
            click.ClickException,
            match=(
                "Component mint was successful but token ID retrieving failed with following error; "
                "Connection interrupted while waiting for the unitId emit event"
            ),
        ):
            with publish_metadata_patch, mock.patch.object(
                registry_contracts, "_component_registry", DummyContract()
            ), mock.patch.object(
                registry_contracts, "_registries_manager", DummyContract()
            ), mock.patch(
                "autonomy.chain.mint.transact"
            ), mock.patch(
                "autonomy.cli.helpers.chain.EthereumApi.try_get_gas_pricing"
            ):
                mint_component(
                    package_path=PACKAGE_DIR,
                    package_type=PackageType.PROTOCOL,
                    key=ETHEREUM_KEY_DEPLOYER,
                    chain_type=ChainType.LOCAL,
                    dependencies=[],
                )

    def test_rpc_cannot_be_none(
        self,
    ) -> None:
        """Test bad chain config."""

        with pytest.raises(
            click.ClickException,
            match=(
                "RPC URL cannot be `None`, "
                "Please set the environment variable for goerli chain "
                "using `GOERLI_CHAIN_RPC` environment variable"
            ),
        ):
            mint_component(
                package_path=PACKAGE_DIR,
                package_type=PackageType.PROTOCOL,
                key=ETHEREUM_KEY_DEPLOYER,
                chain_type=ChainType.GOERLI,
                dependencies=[],
            )

    def test_missing_nft_hash(
        self,
    ) -> None:
        """Test NFT hash not provided failure."""

        with pytest.raises(
            click.ClickException,
            match="Please provide hash for NFT image to mint component on `goerli` chain",
        ):
            with mock.patch(
                "autonomy.cli.helpers.chain.ChainConfigs.get",
                return_value=ChainConfigs.local,
            ):
                mint_component(
                    package_path=PACKAGE_DIR,
                    package_type=PackageType.PROTOCOL,
                    key=ETHEREUM_KEY_DEPLOYER,
                    chain_type=ChainType.GOERLI,
                    dependencies=[],
                )

    def test_mint_component_timeout(
        self,
    ) -> None:
        """Test timeout error."""

        with pytest.raises(
            click.ClickException,
            match=(
                "Component mint was successful but token ID retrieving failed with following error; "
                "Could not retrieve the token in given time limit."
            ),
        ):
            with mock.patch(
                "autonomy.cli.helpers.chain.verify_component_dependencies"
            ), mock.patch(
                "autonomy.cli.helpers.chain.publish_metadata",
                return_value=(None, None),
            ), mock.patch(
                "autonomy.chain.mint.transact"
            ), mock.patch.object(
                registry_contracts._registries_manager,
                "get_create_transaction",
                return_value=False,
            ), mock.patch.object(
                registry_contracts._component_registry,
                "filter_token_id_from_emitted_events",
                return_value=None,
            ):
                mint_component(
                    package_path=PACKAGE_DIR,
                    package_type=PackageType.PROTOCOL,
                    key=ETHEREUM_KEY_DEPLOYER,
                    chain_type=ChainType.LOCAL,
                    dependencies=[],
                    timeout=1.0,
                )


def test_mint_service_timeout() -> None:
    """Test timeout error."""

    with pytest.raises(
        click.ClickException,
        match=(
            "Service mint was successful but token ID retrieving failed with following error; "
            "Could not retrieve the token in given time limit"
        ),
    ):
        with mock.patch(
            "autonomy.cli.helpers.chain.verify_service_dependencies"
        ), mock.patch(
            "autonomy.cli.helpers.chain.load_configuration_object"
        ), mock.patch(
            "autonomy.cli.helpers.chain.publish_metadata",
            return_value=(None, None),
        ), mock.patch(
            "autonomy.chain.mint.transact"
        ), mock.patch.object(
            registry_contracts._service_manager,
            "get_create_transaction",
            return_value=False,
        ), mock.patch.object(
            registry_contracts._service_registry,
            "filter_token_id_from_emitted_events",
            return_value=None,
        ):
            mint_service(
                package_path=PACKAGE_DIR,
                key=ETHEREUM_KEY_DEPLOYER,
                chain_type=ChainType.LOCAL,
                agent_id=1,
                number_of_slots=4,
                cost_of_bond=1,
                threshold=3,
                timeout=1.0,
            )


def test_activate_service_timeout_failure() -> None:
    """Test `activate_service` method"""

    with pytest.raises(
        click.ClickException,
        match="Could not verify the service activation in given time",
    ):
        with mock.patch.object(
            registry_contracts._service_registry,
            "verify_service_has_been_activated",
            return_value=False,
        ), mock.patch.object(
            registry_contracts._service_manager,
            "get_activate_registration_transaction",
            return_value=False,
        ), mock.patch(
            "autonomy.chain.service.transact"
        ), mock.patch(
            "autonomy.chain.service.get_service_info",
            return_value=(1, None, ServiceState.PRE_REGISTRATION.value, None),
        ):
            activate_service(
                service_id=0,
                key=ETHEREUM_KEY_DEPLOYER,
                chain_type=ChainType.LOCAL,
                timeout=1.0,
            )


def test_register_instance_timeout_failure() -> None:
    """Test `deploy_service` method"""

    with pytest.raises(
        click.ClickException,
        match="Could not verify the instance registration for {'0x'} in given time",
    ):
        with mock.patch.object(
            registry_contracts._service_registry,
            "verify_agent_instance_registration",
            return_value=[],
        ), mock.patch.object(
            registry_contracts._service_manager,
            "get_register_instance_transaction",
            return_value=False,
        ), mock.patch(
            "autonomy.chain.service.transact"
        ), mock.patch(
            "autonomy.chain.service.get_service_info",
            return_value=(1, None, ServiceState.ACTIVE_REGISTRATION.value, None),
        ):
            register_instance(
                service_id=0,
                instances=["0x"],
                agent_ids=[1],
                key=ETHEREUM_KEY_DEPLOYER,
                chain_type=ChainType.LOCAL,
                timeout=1.0,
            )


def test_deploy_service_timeout_failure() -> None:
    """Test `deploy_service` method"""

    with pytest.raises(
        click.ClickException,
        match="Could not verify the service deployment for service 0 in given time",
    ):
        with mock.patch.object(
            registry_contracts._service_registry,
            "verify_service_has_been_deployed",
            return_value=False,
        ), mock.patch.object(
            registry_contracts._service_manager,
            "get_service_deploy_transaction",
            return_value=False,
        ), mock.patch(
            "autonomy.chain.service.transact"
        ), mock.patch(
            "autonomy.chain.service.get_service_info",
            return_value=(1, None, ServiceState.FINISHED_REGISTRATION.value, None),
        ):
            deploy_service(
                service_id=0,
                key=ETHEREUM_KEY_DEPLOYER,
                chain_type=ChainType.LOCAL,
                timeout=1.0,
            )


@pytest.mark.parametrize(
    argnames=("state", "error"),
    argvalues=(
        (ServiceState.NON_EXISTENT, "Service does not exist"),
        (ServiceState.PRE_REGISTRATION, "Service not active"),
        (ServiceState.TERMINATED_BONDED, "Service already terminated"),
    ),
)
def test_terminate_service_failures(state: ServiceState, error: str) -> None:
    """Test `terminate_service` method"""

    with pytest.raises(
        click.ClickException,
        match=error,
    ):
        with mock.patch(
            "autonomy.chain.service.get_service_info",
            return_value=(1, None, state.value, None),
        ):
            terminate_service(
                service_id=0,
                key=ETHEREUM_KEY_DEPLOYER,
                chain_type=ChainType.LOCAL,
            )


def test_terminate_service_contract_failure() -> None:
    """Test `terminate_service` method"""

    with pytest.raises(
        click.ClickException,
        match="Service termination failed",
    ), mock.patch.object(
        registry_contracts._service_manager,
        "get_terminate_service_transaction",
        side_effect=ValueError,
    ):
        with mock.patch(
            "autonomy.chain.service.get_service_info",
            return_value=(1, None, ServiceState.FINISHED_REGISTRATION.value, None),
        ):
            terminate_service(
                service_id=0,
                key=ETHEREUM_KEY_DEPLOYER,
                chain_type=ChainType.LOCAL,
            )


@pytest.mark.parametrize(
    argnames=("state", "error"),
    argvalues=(
        (ServiceState.NON_EXISTENT, "Service does not exist"),
        (
            ServiceState.PRE_REGISTRATION,
            "Service needs to be in terminated-bonded state",
        ),
    ),
)
def test_unbond_service_failures(state: ServiceState, error: str) -> None:
    """Test `terminate_service` method"""

    with pytest.raises(
        click.ClickException,
        match=error,
    ):
        with mock.patch(
            "autonomy.chain.service.get_service_info",
            return_value=(1, None, state.value, None),
        ):
            unbond_service(
                service_id=0,
                key=ETHEREUM_KEY_DEPLOYER,
                chain_type=ChainType.LOCAL,
            )


def test_unbond_service_contract_failure() -> None:
    """Test `unbond_service` method"""

    with pytest.raises(
        click.ClickException,
        match="Service unbond failed",
    ), mock.patch.object(
        registry_contracts._service_manager,
        "get_unbond_service_transaction",
        side_effect=ValueError,
    ):
        with mock.patch(
            "autonomy.chain.service.get_service_info",
            return_value=(1, None, ServiceState.TERMINATED_BONDED.value, None),
        ):
            unbond_service(
                service_id=0,
                key=ETHEREUM_KEY_DEPLOYER,
                chain_type=ChainType.LOCAL,
            )


def test_get_ledger_and_crypto_objects() -> None:
    """Test `get_ledger_and_crypto_objects` for hardware wallet support"""

    with mock.patch.object(EthereumHWICrypto, "entity"):
        ledger_api, crypto = get_ledger_and_crypto_objects(
            chain_type=ChainType.LOCAL,
            hwi=True,
        )

    assert isinstance(ledger_api, EthereumHWIApi)
    assert isinstance(crypto, EthereumHWICrypto)


def test_get_ledger_and_crypto_failure() -> None:
    """Test `get_ledger_and_crypto_objects` failures"""

    with pytest.raises(
        click.ClickException,
        match="Please provide key path using `--key` or use `--hwi` if you want to use a hardware wallet",
    ):
        mint_component(
            package_path=PACKAGE_DIR,
            package_type=PackageType.PROTOCOL,
            key=None,
            chain_type=ChainType.LOCAL,
            dependencies=[],
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test mint helpers."""

import re
from pathlib import Path
from typing import Dict
from unittest import mock

import pytest

from autonomy.chain.base import registry_contracts
from autonomy.chain.config import ChainType
from autonomy.chain.constants import COMPONENT_REGISTRY_CONTRACT, CONTRACTS_DIR_LOCAL
from autonomy.chain.exceptions import InvalidMintParameter
from autonomy.chain.mint import mint_service, sort_service_dependency_metadata

from tests.test_autonomy.test_chain.base import DUMMY_HASH


@pytest.mark.parametrize(
    argnames=("parameters", "error_message"),
    argvalues=(
        (
            dict(
                chain_type=ChainType.LOCAL,
                metadata_hash=DUMMY_HASH,
                agent_ids=[],
                number_of_slots_per_agent=[],
                cost_of_bond_per_agent=[],
                threshold=0,
            ),
            "Please provide at least one agent id",
        ),
        (
            dict(
                chain_type=ChainType.LOCAL,
                metadata_hash=DUMMY_HASH,
                agent_ids=[1],
                number_of_slots_per_agent=[],
                cost_of_bond_per_agent=[],
                threshold=0,
            ),
            "Please for provide number of slots for agents",
        ),
        (
            dict(
                chain_type=ChainType.LOCAL,
                metadata_hash=DUMMY_HASH,
                agent_ids=[1],
                number_of_slots_per_agent=[4],
                cost_of_bond_per_agent=[],
                threshold=0,
            ),
            "Please for provide cost of bond for agents",
        ),
        (
            dict(
                chain_type=ChainType.LOCAL,
                metadata_hash=DUMMY_HASH,
                agent_ids=[1],
                number_of_slots_per_agent=[4],
                cost_of_bond_per_agent=[1000],
                threshold=2,
            ),
            re.escape(
                "The threshold value should at least be greater than or equal to ceil((n * 2 + 1) / 3), "
                "n is total number of agent instances in the service"
            ),
        ),
        (
            dict(
                chain_type=ChainType.LOCAL,
                metadata_hash=DUMMY_HASH,
                agent_ids=[1],
                number_of_slots_per_agent=[0],
                cost_of_bond_per_agent=[1000],
                threshold=0,
            ),
            "Number of slots cannot be zero",
        ),
        (
            dict(
                chain_type=ChainType.LOCAL,
                metadata_hash=DUMMY_HASH,
                agent_ids=[1],
                number_of_slots_per_agent=[4],
                cost_of_bond_per_agent=[0],
                threshold=0,
            ),
            "Cost of bond cannot be zero",
        ),
    ),
)
def test_mint_service_invalid_paramters(parameters: Dict, error_message: str) -> None:
    """Test invalid parameters"""

    with pytest.raises(InvalidMintParameter, match=error_message):
        mint_service(
            ledger_api=mock.MagicMock(),  # type: ignore
            crypto=mock.MagicMock(),  # type: ignore
            **parameters
        )


def test_get_contract_method() -> None:
    """Test `get_contract` method"""

    contract = registry_contracts.get_contract(COMPONENT_REGISTRY_CONTRACT)
    assert (
        contract.configuration.directory
        == CONTRACTS_DIR_LOCAL / COMPONENT_REGISTRY_CONTRACT.name
    )

    with mock.patch.object(Path, "exists", return_value=False):
        with pytest.raises(
            FileNotFoundError,
            match=(
                "Contract package valory/component_registry:latest not found in the open-autonomy installation, "
                "please reinstall the package"
            ),
        ):
            contract = registry_contracts.get_contract(COMPONENT_REGISTRY_CONTRACT)


def test_sort_service_dependency_metadata() -> None:
    """Test `sort_service_dependency_metadata` method"""

    agent_ids = [2, 4, 1, 3]
    expected_agent_ids = sorted(agent_ids)

    number_of_slots_per_agents = [8, 4, 2, 1]
    expected_number_of_slots_per_agents = [2, 8, 1, 4]

    cost_of_bond_per_agent = [10, 100, 1000, 10000]
    expected_cost_of_bond_per_agent = [1000, 10, 10000, 100]

    (
        cal_agent_ids,
        cal_number_of_slots_per_agents,
        cal_cost_of_bond_per_agent,
    ) = sort_service_dependency_metadata(
        agent_ids=agent_ids,
        number_of_slots_per_agents=number_of_slots_per_agents,
        cost_of_bond_per_agent=cost_of_bond_per_agent,
    )

    assert cal_agent_ids == expected_agent_ids
    assert cal_number_of_slots_per_agents == expected_number_of_slots_per_agents
    assert cal_cost_of_bond_per_agent == expected_cost_of_bond_per_agent



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Base configurations."""

import os
from collections import OrderedDict
from copy import copy
from typing import Any, Dict, FrozenSet, List, Optional, Sequence, Tuple, cast

from aea.configurations import validation
from aea.configurations.base import (
    ComponentConfiguration,
    ComponentId,
    ConnectionConfig,
    ContractConfig,
)
from aea.configurations.base import (
    PACKAGE_TYPE_TO_CONFIG_CLASS as _PACKAGE_TYPE_TO_CONFIG_CLASS,
)
from aea.configurations.base import PackageConfiguration, ProtocolConfig, SkillConfig
from aea.configurations.data_types import PackageType, PublicId
from aea.exceptions import AEAValidationError
from aea.helpers.base import SimpleIdOrStr
from aea.helpers.env_vars import apply_env_variables, generate_env_vars_recursively

from autonomy.configurations.constants import DEFAULT_SERVICE_CONFIG_FILE, SCHEMAS_DIR
from autonomy.configurations.validation import ConfigValidator


COMPONENT_CONFIGS: Dict = {
    component.package_type.value: component  # type: ignore
    for component in [
        ContractConfig,
        SkillConfig,
        ProtocolConfig,
        ConnectionConfig,
    ]
}


class Service(PackageConfiguration):  # pylint: disable=too-many-instance-attributes
    """Service package configuration."""

    default_configuration_filename = DEFAULT_SERVICE_CONFIG_FILE
    package_type = PackageType.SERVICE
    schema = str(SCHEMAS_DIR.absolute() / "service_schema.json")

    FIELDS_ALLOWED_TO_UPDATE: FrozenSet[str] = frozenset()

    __slots__ = (
        "_name",
        "_author",
        "version",
        "license",
        "fingerprint",
        "fingerprint_ignore_patterns",
        "build_entrypoint",
        "agent",
        "number_of_agents",
        "description",
        "deployment_config",
        "_aea_version",
        "_aea_version_specifiers",
        "_directory",
        "_overrides",
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: SimpleIdOrStr,
        author: SimpleIdOrStr,
        agent: PublicId,
        version: str = "",
        license_: str = "",
        aea_version: str = "",
        fingerprint: Optional[Dict[str, str]] = None,
        fingerprint_ignore_patterns: Optional[Sequence[str]] = None,
        description: str = "",
        number_of_agents: int = 4,
        build_entrypoint: Optional[str] = None,
        overrides: Optional[List] = None,
        deployment: Optional[Dict] = None,
    ) -> None:
        """Initialise object."""

        super().__init__(
            name=name,
            author=author,
            version=version,
            license_=license_,
            aea_version=aea_version,
            build_entrypoint=build_entrypoint,
            fingerprint=fingerprint,
            fingerprint_ignore_patterns=fingerprint_ignore_patterns,
        )

        self.agent = PublicId.from_str(str(agent))
        self.description = description
        self.number_of_agents = number_of_agents
        self.deployment_config = deployment or {}

        self._overrides = [] if overrides is None else overrides

    @property
    def overrides(
        self,
    ) -> List:
        """Returns component overrides."""

        return self._overrides

    @overrides.setter
    def overrides(self, obj: List) -> None:
        """Set overrides."""

        self.check_overrides_valid(obj)
        self._overrides = obj

    @property
    def json(
        self,
    ) -> Dict:
        """Returns an ordered Dict for service config."""

        config = OrderedDict(
            {
                "name": self.name,
                "author": self.author,
                "agent": str(self.agent),
                "version": self.version,
                "license": self.license,
                "aea_version": self.aea_version,
                "description": self.description,
                "number_of_agents": self.number_of_agents,
                "overrides": self.overrides,
                "fingerprint": self.fingerprint,
                "fingerprint_ignore_patterns": self.fingerprint_ignore_patterns,
                "deployment": self.deployment_config,
            }
        )

        return config

    @classmethod
    def _create_or_update_from_json(cls, obj: Dict, instance: Any = None) -> "Service":
        """Create or update from json data."""

        obj = {**(instance.json if instance else {}), **copy(obj)}
        params = dict(
            name=cast(str, obj.get("name")),
            author=cast(str, obj.get("author")),
            agent=cast(str, obj.get("agent")),
            version=cast(str, obj.get("version")),
            license_=cast(str, obj.get("license")),
            aea_version=cast(str, obj.get("aea_version")),
            description=cast(str, obj.get("description")),
            number_of_agents=cast(int, obj.get("number_of_agents")),
            overrides=cast(List, obj.get("overrides", [])),
            fingerprint=cast(Dict[str, str], obj.get("fingerprint", [])),
            fingerprint_ignore_patterns=cast(
                Sequence[str], obj.get("fingerprint_ignore_patterns", [])
            ),
            deployment=obj.get("deployment"),
        )

        return cls(**params)  # type: ignore

    @classmethod
    def validate_config_data(
        cls, json_data: Dict, env_vars_friendly: bool = False
    ) -> None:
        """Validate config data."""
        ConfigValidator(cls.schema, env_vars_friendly=env_vars_friendly).validate(
            json_data
        )

    def check_overrides_valid(
        self, overrides: List, env_vars_friendly: bool = False
    ) -> None:
        """Uses the AEA helper libraries to check individual overrides."""
        base_validator = validation.ConfigValidator("definitions.json")
        processed = []
        for component_configuration_json in overrides:
            configuration, component_id, has_multiple_overrides = self.process_metadata(
                configuration=copy(component_configuration_json)
            )
            if component_id in processed:
                raise AEAValidationError(
                    f"Overrides for component {component_id} are defined more than once"
                )
            if has_multiple_overrides:
                for idx in range(self.number_of_agents):
                    try:
                        _configuration = cast(Dict, configuration[idx])
                    except KeyError as e:
                        raise AEAValidationError(
                            f"Not enough overrides for component {component_id};"
                            f" Number of agents: {self.number_of_agents}"
                        ) from e

                    base_validator.validate_component_configuration(
                        component_id=component_id,
                        configuration=_configuration,
                        env_vars_friendly=env_vars_friendly,
                    )
            else:
                base_validator.validate_component_configuration(
                    component_id=component_id,
                    configuration=configuration,
                    env_vars_friendly=env_vars_friendly,
                )

            processed.append(component_id)

    @staticmethod
    def process_metadata(
        configuration: Dict,
    ) -> Tuple[Dict, ComponentId, bool]:
        """Process component override metadata."""

        component_id = ComponentId(
            component_type=cast(str, configuration.pop("type")),
            public_id=PublicId.from_str(
                cast(
                    str,
                    configuration.pop("public_id"),
                ),
            ),
        )
        _ = configuration.pop("extra", {})
        has_multiple_overrides = all(
            map(lambda x: isinstance(x, int), configuration.keys())
        )
        return configuration, component_id, has_multiple_overrides

    def process_component_overrides(
        self,
        agent_idx: int,
        component_configuration_json: Dict,
    ) -> Dict:
        """
        Process a component configuration in an agent configuration file.

        :param agent_idx: Index of the agent.
        :param component_configuration_json: the JSON object.
        :return: the processed component configuration.
        """

        configuration, component_id, has_multiple_overrides = self.process_metadata(
            configuration=copy(component_configuration_json)
        )

        if has_multiple_overrides:
            configuration = configuration.get(agent_idx, {})
            if configuration == {}:
                raise ValueError(
                    f"Overrides not provided for agent {agent_idx}; component={component_id}"
                )
        configuration = apply_env_variables(
            data=configuration, env_variables=os.environ.copy()
        )
        env_var_dict = self.generate_environment_variables(
            component_id=component_id,
            component_configuration_json=configuration,
        )

        return env_var_dict

    @staticmethod
    def generate_environment_variables(
        component_id: ComponentId,
        component_configuration_json: Dict,
    ) -> Dict:
        """Try to process component with a singular component overrides."""
        config_class = cast(
            ComponentConfiguration,
            COMPONENT_CONFIGS.get(component_id.package_type.value),
        )
        env_var_dict = {}
        export_path_prefix = [
            component_id.package_type.value,
            component_id.name,
        ]
        for field in config_class.FIELDS_ALLOWED_TO_UPDATE:
            field_data = component_configuration_json.get(field, {})
            if field_data == {}:
                continue
            env_var_dict.update(
                generate_env_vars_recursively(
                    data=field_data,
                    export_path=[*export_path_prefix, field],
                )
            )

        return env_var_dict


PACKAGE_TYPE_TO_CONFIG_CLASS = {
    **_PACKAGE_TYPE_TO_CONFIG_CLASS,
    PackageType.SERVICE: Service,
}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helpers for analyse command"""

import re
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple, cast

import click
from aea.components.base import load_aea_package
from aea.configurations.base import AgentConfig, PackageConfiguration, SkillConfig
from aea.configurations.constants import (
    DEFAULT_SKILL_CONFIG_FILE,
    PACKAGE_TYPE_TO_CONFIG_FILE,
)
from aea.configurations.data_types import (
    ComponentId,
    ComponentType,
    PackageId,
    PackageType,
    PublicId,
)
from aea.configurations.loader import load_configuration_object
from aea.crypto.base import LedgerApi
from aea.package_manager.v1 import PackageManagerV1
from aea_cli_ipfs.ipfs_utils import IPFSTool

from autonomy.analyse.dialogues import check_dialogues_in_a_skill_package
from autonomy.analyse.logs.base import (
    ENTER_ROUND_REGEX,
    EXIT_ROUND_REGEX,
    LOGS_DB,
    LogRow,
)
from autonomy.analyse.logs.collection import FromDirectory, LogCollection
from autonomy.analyse.logs.db import AgentLogsDB
from autonomy.analyse.service import ServiceAnalyser, ServiceValidationFailed
from autonomy.chain.config import ChainType, ContractConfigs
from autonomy.chain.exceptions import FailedToRetrieveComponentMetadata
from autonomy.chain.utils import resolve_component_id
from autonomy.cli.helpers.chain import get_ledger_and_crypto_objects
from autonomy.cli.utils.click_utils import sys_path_patch
from autonomy.configurations.base import PACKAGE_TYPE_TO_CONFIG_CLASS, Service


def load_package_tree(packages_dir: Path) -> None:
    """Load package tree."""

    pm = PackageManagerV1.from_dir(packages_dir=packages_dir)

    for package_id in pm.iter_dependency_tree():
        if package_id.package_type in (PackageType.AGENT, PackageType.SERVICE):
            continue

        package_path = pm.package_path_from_package_id(package_id=package_id)
        config_obj = load_configuration_object(
            package_type=package_id.package_type, directory=package_path
        )
        config_obj.directory = package_path
        load_aea_package(configuration=config_obj)


def list_all_skill_yaml_files(registry_path: Path) -> List[Path]:
    """List all skill yaml files in a local registry"""
    return sorted(registry_path.glob(f"*/skills/*/{DEFAULT_SKILL_CONFIG_FILE}"))


def run_dialogues_check(
    packages_dir: Path,
    ignore: List[str],
    dialogues: List[str],
) -> None:
    """Run dialogues check."""

    try:
        with sys_path_patch(packages_dir.parent):
            load_package_tree(packages_dir=packages_dir)
            for yaml_file in list_all_skill_yaml_files(packages_dir):
                if yaml_file.parent.name in ignore:
                    click.echo(f"Skipping {yaml_file.parent.name}")
                    continue

                click.echo(f"Checking {yaml_file.parent.name}")
                check_dialogues_in_a_skill_package(
                    config_file=yaml_file,
                    dialogues=dialogues,
                )
    except (ValueError, FileNotFoundError, ImportError) as e:
        raise click.ClickException(str(e))


class ParseLogs:
    """Parse agent logs."""

    _dbs: Dict[str, AgentLogsDB]
    _collection: LogCollection
    _db_path: Path

    results: Dict[str, List[LogRow]]

    def __init__(self) -> None:
        """Initialize object."""

    @property
    def agents(self) -> List[str]:
        """Available agents."""

        return self._collection.agents

    @property
    def n_agents(self) -> int:
        """Available agents."""

        return self._collection.n_agents

    def from_dir(self, logs_dir: Path) -> "ParseLogs":
        """From directory"""

        self._db_path = logs_dir / LOGS_DB
        self._collection = FromDirectory(directory=logs_dir)
        self._dbs = {
            agent: AgentLogsDB(agent=agent, file=self._db_path)
            for agent in self._collection.agents
        }

        return self

    def create_tables(self, reset: bool = False) -> "ParseLogs":
        """Create required tables."""

        for agent, db in self._dbs.items():
            db_exists = db.exists()
            if db_exists and not reset:
                continue

            self._collection.create_agent_db(
                agent=agent,
                db=db,
                reset=reset,
            )

        return self

    def select(  # pylint: disable=too-many-arguments
        self,
        agents: List[str],
        start_time: Optional[datetime],
        end_time: Optional[datetime],
        log_level: Optional[str],
        period: Optional[int],
        round_name: Optional[str],
        behaviour_name: Optional[str],
    ) -> "ParseLogs":
        """Query and return results."""

        results = {}
        for agent in agents:
            results[agent] = self._dbs[agent].select(
                start_time=start_time,
                end_time=end_time,
                log_level=log_level,
                period=period,
                round_name=round_name,
                behaviour_name=behaviour_name,
            )

        self.results = results
        return self

    def re_include(self, regexes: List[str]) -> "ParseLogs":
        """Apply a set of regexes on the result."""
        if len(regexes) == 0:
            return self

        compiled_re = [re.compile(pattern) for pattern in regexes]

        def _apply_filter(row: LogRow) -> bool:
            return any(map(lambda _re: _re.match(row[2]) is not None, compiled_re))

        for agent, logs in self.results.items():
            self.results[agent] = list(filter(_apply_filter, logs))

        return self

    def re_exclude(self, regexes: List[str]) -> "ParseLogs":
        """Apply a set of regexes on the result."""

        if len(regexes) == 0:
            return self

        compiled_re = [re.compile(pattern) for pattern in regexes]

        def _apply_filter(row: LogRow) -> bool:
            return any(map(lambda _re: _re.match(row[2]) is None, compiled_re))

        for agent, logs in self.results.items():
            self.results[agent] = list(filter(_apply_filter, logs))

        return self

    def execution_path(self) -> None:
        """Output FSM path"""
        for agent, logs in self.results.items():
            period = -1
            click.echo(f"Agent {agent}")
            for _, _, message, _, _, _ in logs:
                match = ENTER_ROUND_REGEX.match(message)
                if match is not None:
                    _, _period = cast(Tuple[str, int], match.groups())
                    if _period != period:
                        period = _period
                        click.echo(f"|_ Period {period}")

                match = EXIT_ROUND_REGEX.match(message)
                if match is not None:
                    round_name, exit_event = match.groups()
                    click.echo(f"| |_ {round_name} | {exit_event}")
            click.echo("|_ End\n")

    def table(self) -> None:
        """Print table."""

        for agent, logs in self.results.items():
            click.echo(f"--- {agent} ---")
            for timestamp, log_level, message, _, _, _ in logs:
                click.echo(f"[{timestamp}][{log_level}] {message}")
            click.echo("--- End ---")


def _get_content_from_ipfs(package_id: PackageId, file: str) -> bytes:
    """Read content from the IPFS registry."""
    try:
        ipfs_tool = IPFSTool()
        return ipfs_tool.client.cat(
            f"{package_id.package_hash}/{package_id.name}/{file}"
        )
    except Exception as e:
        message, *_ = e.args
        if "merkledag: not found" in message:
            message = "File does not exist on the IPFS registry"

        raise click.ClickException(
            "Fetching content from the IPFS registry failed"
            f"\n\tPackage: {package_id}\n\tFile: {file}\n\tError: {message}"
        ) from e


def _load_from_ipfs(package_id: PackageId) -> PackageConfiguration:
    """Load agent config from the IPFS hash."""
    with tempfile.TemporaryDirectory() as package_dir:
        config_file = PACKAGE_TYPE_TO_CONFIG_FILE[package_id.package_type.value]
        config_content = _get_content_from_ipfs(package_id=package_id, file=config_file)
        Path(package_dir, config_file).write_bytes(config_content)
        return load_configuration_object(
            package_type=package_id.package_type,
            directory=Path(package_dir),
            package_type_config_class=PACKAGE_TYPE_TO_CONFIG_CLASS,
        )


def _load_from_local(
    package_id: PackageId, package_manager: PackageManagerV1
) -> PackageConfiguration:
    """Load from local repository."""

    return load_configuration_object(
        package_type=package_id.package_type,
        directory=package_manager.package_path_from_package_id(package_id=package_id),
        package_type_config_class=PACKAGE_TYPE_TO_CONFIG_CLASS,
    )


def _get_chained_abci_skill(
    agent_config: AgentConfig,
    package_manager: PackageManagerV1,
    is_on_chain_check: bool = False,
) -> Optional[SkillConfig]:
    """
    Get chained ABCI skill config

    This method cycles through the component overrides defined inside an agent package
    and tries to find out the chained ABCI app.

    :param agent_config: Agent configuration object.
    :param package_manager: Package manager.
    :param is_on_chain_check: A boolean flag to specify whether this is an on-chain
                            check or a local check
    :return: Skill configuration object if found
    """

    for skill_id in agent_config.skills:
        override = agent_config.component_configurations.get(
            ComponentId(component_type=ComponentType.SKILL, public_id=skill_id)
        )
        if override is None:
            continue

        package_id = PackageId(
            package_type=PackageType.SKILL,
            public_id=skill_id,
        )

        skill_config = cast(
            SkillConfig,
            (
                _load_from_ipfs(package_id=package_id)
                if is_on_chain_check
                else _load_from_local(
                    package_id=package_id, package_manager=package_manager
                )
            ),
        )

        # Check if the skill override has the `is_abstract` property set to true
        if override.get("is_abstract", False):
            continue

        # Check if the skill has the `is_abstract` property set to true
        if skill_config.is_abstract:
            continue

        # This statement makes an assumption skills other than the chained/main
        # abci are defined as abstract
        return skill_config

    return None


def _get_ipfs_pins(is_on_chain_check: bool = False) -> Set[str]:
    """Get IPFS pins."""
    if is_on_chain_check:
        ipfs_tool = IPFSTool()
        if not ipfs_tool.daemon.is_started():
            raise click.ClickException("Cannot connect to the IPFS daemon.")
        return ipfs_tool.all_pins()

    return set()


def _get_on_chain_service_id(
    ledger_api: LedgerApi,
    chain_type: ChainType,
    token_id: int,
) -> PackageId:
    """Get on chain service ID with package hash"""
    service_info = resolve_component_id(
        ledger_api=ledger_api,
        contract_address=ContractConfigs.service_registry.contracts[chain_type],
        token_id=cast(int, token_id),
        is_service=True,
    )
    package_hash = cast(str, service_info["code_uri"]).replace("ipfs://", "")
    public_id = PublicId.from_str(service_info["name"]).with_hash(
        package_hash=package_hash
    )
    return PackageId(package_type=PackageType.SERVICE, public_id=public_id)


def check_service_readiness(  # pylint: disable=too-many-locals
    token_id: Optional[int],
    public_id: Optional[PublicId],
    chain_type: ChainType,
    packages_dir: Path,
    skip_warnings: bool = False,
) -> None:
    """Check deployment readiness of a service."""

    is_on_chain_check = token_id is not None
    ledger_api, _ = get_ledger_and_crypto_objects(chain_type=chain_type)
    package_manager = PackageManagerV1.from_dir(packages_dir=packages_dir)
    ipfs_pins = _get_ipfs_pins(is_on_chain_check=is_on_chain_check)

    if is_on_chain_check:
        try:
            service_id = _get_on_chain_service_id(
                ledger_api=ledger_api,
                chain_type=chain_type,
                token_id=cast(int, token_id),
            )
        except FailedToRetrieveComponentMetadata as e:
            raise click.ClickException(
                f"On chain service verification failed; {e}"
            ) from e
    else:
        service_id = PackageId(package_type=PackageType.SERVICE, public_id=public_id)

    service_config = cast(
        Service,
        (
            _load_from_ipfs(package_id=service_id)
            if is_on_chain_check
            else _load_from_local(
                package_id=service_id, package_manager=package_manager
            )
        ),
    )

    agent_id = PackageId(
        package_type=PackageType.AGENT,
        public_id=service_config.agent,
    )
    agent_config = cast(
        AgentConfig,
        (
            _load_from_ipfs(package_id=agent_id)
            if is_on_chain_check
            else _load_from_local(package_id=agent_id, package_manager=package_manager)
        ),
    )

    skill_config = _get_chained_abci_skill(
        agent_config=agent_config,
        package_manager=package_manager,
        is_on_chain_check=is_on_chain_check,
    )

    if skill_config is None:
        raise click.ClickException(
            "Please make sure the agent package configuration contains overrides for the chained ABCI app"
        )

    try:
        service_analyser = ServiceAnalyser(
            service_config=service_config,
            abci_skill_id=skill_config.public_id,
            is_on_chain_check=is_on_chain_check,
            skip_warnings=skip_warnings,
        )

        service_analyser.check_on_chain_state(
            ledger_api=ledger_api,
            chain_type=chain_type,
            token_id=cast(int, token_id),
        )
        service_analyser.validate_skill_config(skill_config=skill_config)
        service_analyser.validate_agent_overrides(agent_config=agent_config)
        service_analyser.validate_agent_override_env_vars(agent_config=agent_config)
        service_analyser.validate_service_overrides()
        service_analyser.validate_service_override_env_vars()
        service_analyser.cross_verify_overrides(
            agent_config=agent_config,
            skill_config=skill_config,
        )
        service_analyser.check_agent_dependencies_published(
            ipfs_pins=ipfs_pins, agent_config=agent_config
        )

        click.echo("Service is ready to be deployed.")
    except ServiceValidationFailed as e:
        raise click.ClickException(str(e)) from e
